@@ -78,6 +78,15 @@ use_copyplane(PixmapPtr dst, GCPtr gc, glamor_program *prog, void *arg)
 
     /* XXX handle 2 10 10 10 and 1555 formats; presumably the pixmap private knows this? */
     switch (args->src_pixmap->drawable.depth) {
+    case 30:
+        glUniform4ui(prog->bitplane_uniform,
+                     (args->bitplane >> 20) & 0x3ff,
+                     (args->bitplane >> 10) & 0x3ff,
+                     (args->bitplane      ) & 0x3ff,
+                     0);
+
+        glUniform4f(prog->bitmul_uniform, 0x3ff, 0x3ff, 0x3ff, 0);
+        break;
     case 24:
         glUniform4ui(prog->bitplane_uniform,
                      (args->bitplane >> 16) & 0xff,
@@ -180,7 +189,7 @@ glamor_copy_bail(DrawablePtr src,
 }
 
 /**
- * Implements CopyPlane and CopyArea from the GPU to the GPU by using
+ * Implements CopyPlane and CopyArea from the CPU to the GPU by using
  * the source as a texture and painting that into the destination.
  *
  * This requires that source and dest are different textures, or that
@@ -203,10 +212,6 @@ glamor_copy_cpu_fbo(DrawablePtr src,
     ScreenPtr screen = dst->pScreen;
     glamor_screen_private *glamor_priv = glamor_get_screen_private(screen);
     PixmapPtr dst_pixmap = glamor_get_drawable_pixmap(dst);
-    FbBits *src_bits;
-    FbStride src_stride;
-    int src_bpp;
-    int src_xoff, src_yoff;
     int dst_xoff, dst_yoff;
 
     if (gc && gc->alu != GXcopy)
@@ -221,33 +226,43 @@ glamor_copy_cpu_fbo(DrawablePtr src,
     glamor_get_drawable_deltas(dst, dst_pixmap, &dst_xoff, &dst_yoff);
 
     if (bitplane) {
-        PixmapPtr src_pix = fbCreatePixmap(screen, dst_pixmap->drawable.width,
+        FbBits *tmp_bits;
+        FbStride tmp_stride;
+        int tmp_bpp;
+        int tmp_xoff, tmp_yoff;
+
+        PixmapPtr tmp_pix = fbCreatePixmap(screen, dst_pixmap->drawable.width,
                                            dst_pixmap->drawable.height,
                                            dst->depth, 0);
 
-        if (!src_pix) {
+        if (!tmp_pix) {
             glamor_finish_access(src);
             goto bail;
         }
 
-        src_pix->drawable.x = -dst->x;
-        src_pix->drawable.y = -dst->y;
+        tmp_pix->drawable.x = dst_xoff;
+        tmp_pix->drawable.y = dst_yoff;
 
-        fbGetDrawable(&src_pix->drawable, src_bits, src_stride, src_bpp, src_xoff,
-                      src_yoff);
+        fbGetDrawable(&tmp_pix->drawable, tmp_bits, tmp_stride, tmp_bpp, tmp_xoff,
+                      tmp_yoff);
 
         if (src->bitsPerPixel > 1)
-            fbCopyNto1(src, &src_pix->drawable, gc, box, nbox, dx, dy,
+            fbCopyNto1(src, &tmp_pix->drawable, gc, box, nbox, dx, dy,
                        reverse, upsidedown, bitplane, closure);
         else
-            fbCopy1toN(src, &src_pix->drawable, gc, box, nbox, dx, dy,
+            fbCopy1toN(src, &tmp_pix->drawable, gc, box, nbox, dx, dy,
                        reverse, upsidedown, bitplane, closure);
 
-        glamor_upload_boxes(dst_pixmap, box, nbox, src_xoff, src_yoff,
-                            dst_xoff, dst_yoff, (uint8_t *) src_bits,
-                            src_stride * sizeof(FbBits));
-        fbDestroyPixmap(src_pix);
+        glamor_upload_boxes(dst_pixmap, box, nbox, tmp_xoff, tmp_yoff,
+                            dst_xoff, dst_yoff, (uint8_t *) tmp_bits,
+                            tmp_stride * sizeof(FbBits));
+        fbDestroyPixmap(tmp_pix);
     } else {
+        FbBits *src_bits;
+        FbStride src_stride;
+        int src_bpp;
+        int src_xoff, src_yoff;
+
         fbGetDrawable(src, src_bits, src_stride, src_bpp, src_xoff, src_yoff);
         glamor_upload_boxes(dst_pixmap, box, nbox, src_xoff + dx, src_yoff + dy,
                             dst_xoff, dst_yoff,
@@ -311,6 +326,13 @@ bail:
     return FALSE;
 }
 
+/* Include the enums here for the moment, to keep from needing to bump epoxy. */
+#ifndef GL_TILE_RASTER_ORDER_FIXED_MESA
+#define GL_TILE_RASTER_ORDER_FIXED_MESA          0x8BB8
+#define GL_TILE_RASTER_ORDER_INCREASING_X_MESA   0x8BB9
+#define GL_TILE_RASTER_ORDER_INCREASING_Y_MESA   0x8BBA
+#endif
+
 /*
  * Copy from GPU to GPU by using the source
  * as a texture and painting that into the destination
@@ -344,6 +366,8 @@ glamor_copy_fbo_fbo_draw(DrawablePtr src,
     glamor_program *prog;
     const glamor_facet *copy_facet;
     int n;
+    Bool ret = FALSE;
+    BoxRec bounds = glamor_no_rendering_bounds();
 
     glamor_make_current(glamor_priv);
 
@@ -380,15 +404,34 @@ glamor_copy_fbo_fbo_draw(DrawablePtr src,
 
     v = glamor_get_vbo_space(dst->pScreen, nbox * 8 * sizeof (int16_t), &vbo_offset);
 
+    if (src_pixmap == dst_pixmap && glamor_priv->has_mesa_tile_raster_order) {
+        glEnable(GL_TILE_RASTER_ORDER_FIXED_MESA);
+        if (dx >= 0)
+            glEnable(GL_TILE_RASTER_ORDER_INCREASING_X_MESA);
+        else
+            glDisable(GL_TILE_RASTER_ORDER_INCREASING_X_MESA);
+        if (dy >= 0)
+            glEnable(GL_TILE_RASTER_ORDER_INCREASING_Y_MESA);
+        else
+            glDisable(GL_TILE_RASTER_ORDER_INCREASING_Y_MESA);
+    }
+
     glEnableVertexAttribArray(GLAMOR_VERTEX_POS);
     glVertexAttribPointer(GLAMOR_VERTEX_POS, 2, GL_SHORT, GL_FALSE,
                           2 * sizeof (GLshort), vbo_offset);
 
+    if (nbox < 100) {
+        bounds = glamor_start_rendering_bounds();
+        for (int i = 0; i < nbox; i++)
+            glamor_bounds_union_box(&bounds, &box[i]);
+    }
+
     for (n = 0; n < nbox; n++) {
         v[0] = box->x1; v[1] = box->y1;
         v[2] = box->x1; v[3] = box->y2;
         v[4] = box->x2; v[5] = box->y2;
         v[6] = box->x2; v[7] = box->y1;
+
         v += 8;
         box++;
     }
@@ -410,25 +453,39 @@ glamor_copy_fbo_fbo_draw(DrawablePtr src,
             goto bail_ctx;
 
         glamor_pixmap_loop(dst_priv, dst_box_index) {
-            glamor_set_destination_drawable(dst, dst_box_index, FALSE, FALSE,
-                                            prog->matrix_uniform,
-                                            &dst_off_x, &dst_off_y);
+            BoxRec scissor = {
+                .x1 = max(-args.dx, bounds.x1),
+                .y1 = max(-args.dy, bounds.y1),
+                .x2 = min(-args.dx + src_box->x2 - src_box->x1, bounds.x2),
+                .y2 = min(-args.dy + src_box->y2 - src_box->y1, bounds.y2),
+            };
+            if (scissor.x1 >= scissor.x2 || scissor.y1 >= scissor.y2)
+                continue;
 
-            glScissor(dst_off_x - args.dx,
-                      dst_off_y - args.dy,
-                      src_box->x2 - src_box->x1,
-                      src_box->y2 - src_box->y1);
+            if (!glamor_set_destination_drawable(dst, dst_box_index, FALSE, FALSE,
+                                                 prog->matrix_uniform,
+                                                 &dst_off_x, &dst_off_y))
+                goto bail_ctx;
+
+            glScissor(scissor.x1 + dst_off_x,
+                      scissor.y1 + dst_off_y,
+                      scissor.x2 - scissor.x1,
+                      scissor.y2 - scissor.y1);
 
             glamor_glDrawArrays_GL_QUADS(glamor_priv, nbox);
         }
     }
+
+    ret = TRUE;
+
+bail_ctx:
+    if (src_pixmap == dst_pixmap && glamor_priv->has_mesa_tile_raster_order) {
+        glDisable(GL_TILE_RASTER_ORDER_FIXED_MESA);
+    }
     glDisable(GL_SCISSOR_TEST);
     glDisableVertexAttribArray(GLAMOR_VERTEX_POS);
 
-    return TRUE;
-
-bail_ctx:
-    return FALSE;
+    return ret;
 }
 
 /**
@@ -585,34 +642,36 @@ glamor_copy_needs_temp(DrawablePtr src,
     if (!glamor_priv->has_nv_texture_barrier)
         return TRUE;
 
-    glamor_get_drawable_deltas(src, src_pixmap, &src_off_x, &src_off_y);
-    glamor_get_drawable_deltas(dst, dst_pixmap, &dst_off_x, &dst_off_y);
+    if (!glamor_priv->has_mesa_tile_raster_order) {
+        glamor_get_drawable_deltas(src, src_pixmap, &src_off_x, &src_off_y);
+        glamor_get_drawable_deltas(dst, dst_pixmap, &dst_off_x, &dst_off_y);
 
-    bounds = box[0];
-    for (n = 1; n < nbox; n++) {
-        bounds.x1 = min(bounds.x1, box[n].x1);
-        bounds.y1 = min(bounds.y1, box[n].y1);
+        bounds = box[0];
+        for (n = 1; n < nbox; n++) {
+            bounds.x1 = min(bounds.x1, box[n].x1);
+            bounds.y1 = min(bounds.y1, box[n].y1);
 
-        bounds.x2 = max(bounds.x2, box[n].x2);
-        bounds.y2 = max(bounds.y2, box[n].y2);
-    }
+            bounds.x2 = max(bounds.x2, box[n].x2);
+            bounds.y2 = max(bounds.y2, box[n].y2);
+        }
 
-    /* Check to see if the pixmap-relative boxes overlap in both X and Y,
-     * in which case we can't rely on NV_texture_barrier and must
-     * make a temporary copy
-     *
-     *  dst.x1                     < src.x2 &&
-     *  src.x1                     < dst.x2 &&
-     *
-     *  dst.y1                     < src.y2 &&
-     *  src.y1                     < dst.y2
-     */
-    if (bounds.x1 + dst_off_x      < bounds.x2 + dx + src_off_x &&
-        bounds.x1 + dx + src_off_x < bounds.x2 + dst_off_x &&
+        /* Check to see if the pixmap-relative boxes overlap in both X and Y,
+         * in which case we can't rely on NV_texture_barrier and must
+         * make a temporary copy
+         *
+         *  dst.x1                     < src.x2 &&
+         *  src.x1                     < dst.x2 &&
+         *
+         *  dst.y1                     < src.y2 &&
+         *  src.y1                     < dst.y2
+         */
+        if (bounds.x1 + dst_off_x      < bounds.x2 + dx + src_off_x &&
+            bounds.x1 + dx + src_off_x < bounds.x2 + dst_off_x &&
 
-        bounds.y1 + dst_off_y      < bounds.y2 + dy + src_off_y &&
-        bounds.y1 + dy + src_off_y < bounds.y2 + dst_off_y) {
-        return TRUE;
+            bounds.y1 + dst_off_y      < bounds.y2 + dy + src_off_y &&
+            bounds.y1 + dy + src_off_y < bounds.y2 + dst_off_y) {
+            return TRUE;
+        }
     }
 
     glTextureBarrierNV();
